// pvi.c: Signetics 2636 PVI emulation module (platform-independent)

// #define DEBUGSPRITES
// #define DEBUGINTERRUPTS
// #define DEBUGCOLLISIONS
// note that DEBUGCOLLISIONS only debugs sprite-sprite collisions, not sprite-background collisions
// #define NOSPRITES
// #define NOGRID
// #define SELECTIVESPRITES
// allows you to selectively disable sprites

#define RASTLINES 312

/* This is how each frame is currently emulated:

now VRST is turned off
  0..  19 are  20 "margin" (above grid but visible) Y-pixels
  20..219 are 200 "grid"                            Y-pixels
 220..271 are  52 "margin" (below grid but visible) Y-pixels
now VRST is turned on
 272..291 are  20 "vertical retrace" (below grid)   Y-pixels
the true frame boundary is here
 292..311 are  20 "vertical retrace" (above grid)   Y-pixels

Almost but not quite the same as shown in the Interton VC 4000 Coding
 Guide. Doing it the same way as the Coding Guide doesn't seem to help.

Interton VC 4000: assumed behaviour is that the low nybbles of CONSOLE and
PnFOOKEYS are set to %1111, the same as is proven to occur on an Elektor
TVGC. */

#ifdef AMIGA
    #include "lint.h"
    #include "amiga.h"

    #include <proto/locale.h> // GetCatalogStr()
#endif
#ifdef WIN32
    #include "ibm.h"
    #define EXEC_TYPES_H
#endif

#include "aa.h"
#define CATCOMP_NUMBERS
#define CATCOMP_CODE
#define CATCOMP_BLOCK
#include "aa_strings.h"

#include <stdio.h>
#include <string.h>

#define I_XSIZE 227
#define HIDDEN_X  5 /* opposite of XMARGIN
(ie. Arcadia emulator has overscan but Interton/Elektor emulator has
underscan)
Circledrive   (Elektor) prefers a value of >=  9 for this.
Painting      (Elektor) prefers a value of >=  9 for this.
Tiny Tim      (Elektor) prefers a value of >=  9 for this.
Omega Landing (Elektor) prefers a value of >= 15 for this. */

EXPORT UWORD ie_spritedata[4] = { I_SPRITE0,
                                  I_SPRITE1,
                                  I_SPRITE2,
                                  I_SPRITE3 };

#ifdef SELECTIVESPRITES
    MODULE FLAG enabled[4] = {TRUE, TRUE, TRUE, TRUE};
#endif

MODULE struct
{   int   delay,
          state,
          x;
    UBYTE y;// MUST be UBYTE!
    UBYTE colour;
} sprite[4];

MODULE int    fgc,
              whichsprite,
              x,
              y,
              xx;
MODULE UBYTE  digitcolour,
              OutputBuffer[7],
              t;
MODULE FLAG   rumble1, rumble2;

IMPORT FLAG   forceinterrupt;
IMPORT UBYTE  screen[BOXWIDTH][BOXHEIGHT];
IMPORT ULONG  analog,
              autofire,
              collisions,
              flagline,
              paused,
              swapped,
              warp;
IMPORT UBYTE* IOBuffer;
IMPORT FILE*  MacroHandle;
IMPORT TEXT   friendly[2 + 13 + 1 + 1];
IMPORT int    connected,
              cpl,
              fudge,
              frameskip,
              joy1,
              key1,
              key2,
              key3,
              key4,
              p1sprite,
              p2sprite,
              machine,
              memmap,
              offset,
              overcalc, // must be signed!
              rast,
              rastn,
              recmode,
              recsize,
              requirebutton,
              sensitivity,
              slice,
              wp,
              consoleopen,
              crippled,
              inframe,
              interrupt,
              limitrefresh,
              loop,
              runtorastline,
              runtoframe,
              trace;
IMPORT SLONG  ax[2], // analog paddle X-coords
              ay[2]; // analog paddle Y-coords
IMPORT UBYTE  psu,
              memory[32768],
              sx[2], sy[2];
IMPORT UWORD  console[4],
              joystate,
              keypads[2][NUMKEYS],
              page,
              iar;
IMPORT ULONG  downframes,
              elapsed,
              frames,
              totalframes;
IMPORT struct KeyTableStruct keytable[16];
IMPORT FLAG                  absolute;

#ifdef WIN32
    IMPORT ULONG*          display;
    IMPORT ULONG           pens[16];
    IMPORT int             CatalogPtr;
#endif
#ifdef AMIGA
    IMPORT LONG            pens[16],
                           wordpens[16],
                           longpens[16];
    IMPORT int             divided,
                           fullscreen,
                        // remainder,
                           px, py,
                           wide,
                           size;
    IMPORT UBYTE*          byteptr[MAXBOXHEIGHT];
    IMPORT UWORD*          wordptr[MAXBOXHEIGHT];
    IMPORT ULONG*          longptr[MAXBOXHEIGHT];
    IMPORT struct Catalog* CatalogPtr;
#endif

MODULE __inline void drawgrid(void);
MODULE void drawdigit(int x, int y, int d, int line);
MODULE __inline void drawsprites(void);
MODULE void ie_playerinput(int source, int dest);
MODULE __inline void ie_emuinput(void);
MODULE __inline void breakrastline(void);
MODULE __inline void ie_drawpixel(int x, int y, int colour);
MODULE __inline void ie_drawline(int colour);
MODULE __inline UBYTE pviread(int address);

// Each "character" is 12*20 pixels (3*5 "chunky 4*4" pixels).
MODULE const char led[20][12 + 1] =
{   "aaaabbbbcccc",
    "aaaabbbbcccc",
    "aaaabbbbcccc",
    "aaaabbbbcccc",
    "llll    dddd",
    "llll    dddd",
    "llll    dddd",
    "llll    dddd",
    "kkkkmmmmeeee",
    "kkkkmmmmeeee",
    "kkkkmmmmeeee",
    "kkkkmmmmeeee",
    "jjjj    ffff",
    "jjjj    ffff",
    "jjjj    ffff",
    "jjjj    ffff",
    "iiiihhhhgggg",
    "iiiihhhhgggg",
    "iiiihhhhgggg",
    "iiiihhhhgggg"
};

MODULE int colltable[I_XSIZE];
/* bits are:
   7..4: sprites #3..#0 respectively
   3:    grid
   2..0: sprite foreground colour (wire OR'ed) */

EXPORT void pvi(void)
{   // assert(machine != ARCADIA);

    inframe = TRUE;
    rumble1 = rumble2 = FALSE;

    // beginning of frame?

    pviwrite(I_BGCOLLIDE ,     0); // clear sprite complete and sprite-bg collision bits
    pviwrite(I_SPRITECOLLIDE , 0); // clear VRST and sprite-sprite collision bits
    pviwrite(I_BGCOLLIDE2,     0); // clear sprite complete and sprite-bg collision bits
    pviwrite(I_SPRITECOLLIDE2, 0); // clear VRST and sprite-sprite collision bits
    pviwrite(I_BGCOLLIDE3,     0); // clear sprite complete and sprite-bg collision bits
    pviwrite(I_SPRITECOLLIDE3, 0); // clear VRST and sprite-sprite collision bits
    pviwrite(I_BGCOLLIDE4,     0); // clear sprite complete and sprite-bg collision bits
    pviwrite(I_SPRITECOLLIDE4, 0); // clear VRST and sprite-sprite collision bits

    psu &= ~(PSU_S);
    ie_emuinput();

    for (whichsprite = 0; whichsprite < 4; whichsprite++)
    {   sprite[whichsprite].y     = pviread(ie_spritedata[whichsprite] + 12); // SPRITEnAY
        sprite[whichsprite].y++; // overflow OK!
        sprite[whichsprite].state = 0;
    }
    for (rast = 0; rast < RASTLINES; rast++)
    {   breakrastline();

        slice = cpl - overcalc;
        cpu();
        CHECKINPUT;

        // blank the current rastline
        // only doing this for rast < IE_BOXHEIGHT isn't good enough;
        // it needs to be done for every rastline
        for (x = 0; x < I_XSIZE; x++)
        {   colltable[x] = 0;
        }

        if (rast < IE_BOXHEIGHT)
        {   drawgrid();

            if
            (   (rast >=  20 && rast <=  39)
             || (rast >= 200 && rast <= 219)
            )
            {   if (pviread(I_SCORECTRL) & 1)
                {   y = 200;
                } else
                {   y =  20;
                }
                if (rast == y)
                {   if (pviread(I_BGCOLOUR) & 8)
                    {   digitcolour = (pviread(I_BGCOLOUR) >> 4) & 7;
                    } else
                    {   digitcolour = 0; // white
                }   }
                // this is cached for eg. LASERATT game over screen
                if (rast >= y && rast < y + 20) // 20 rastlines is the height of each digit
                {   drawdigit(    32 - HIDDEN_X + 28, y, pviread(I_SCORELT) >>   4, rast - y);
                    drawdigit(    32 - HIDDEN_X + 44, y, pviread(I_SCORELT) &  0xf, rast - y);
                    if (pviread(I_SCORECTRL) & 2)
                    {   drawdigit(32 - HIDDEN_X + 60, y, pviread(I_SCORERT) >>   4, rast - y);
                        drawdigit(32 - HIDDEN_X + 76, y, pviread(I_SCORERT) &  0xf, rast - y);
                    } else
                    {   drawdigit(32 - HIDDEN_X + 76, y, pviread(I_SCORERT) >>   4, rast - y);
                        drawdigit(32 - HIDDEN_X + 92, y, pviread(I_SCORERT) &  0xf, rast - y);
        }   }   }   }

        // sprites raster must be drawn AFTER the background raster

        drawsprites();

        // Once a collision is registered, does it remain set
        // for the remainder of the frame, or only the current
        // rastline? (We're assuming it's for the remainder of
        // the frame.)

        if (collisions)
        {   t = 0;
            /* ESS-011-7-OmegaLanding and ESS-011-D-Painting seem to expect collisions at the
               far left (< 9) to not register. */
            for (x = 9; x < I_XSIZE; x++) // bitwise...       3210,g,fgc
            {   if ((colltable[x] & 0x30) == 0x30)
                {   t |= 32; // sprites #0 and #1 collision (%0011,0,000)
#ifdef DEBUGCOLLISIONS
                    T1; printf("Sprites #0 and #1 have collided at pixel %ld!\n", x); T2;
#endif
                    if (p1sprite == 0 || p1sprite == 1)
                    {   rumble1 = TRUE;
                    }
                    if (p2sprite == 0 || p2sprite == 1)
                    {   rumble2 = TRUE;
                }   }
                if ((colltable[x] & 0x50) == 0x50)
                {   t |= 16; // sprites #0 and #2 collision (%0101,0,000)
#ifdef DEBUGCOLLISIONS
                    T1; printf("Sprites #0 and #2 have collided at pixel %ld!\n", x); T2;
#endif
                    if (p1sprite == 0 || p1sprite == 2)
                    {   rumble1 = TRUE;
                    }
                    if (p2sprite == 0 || p2sprite == 2)
                    {   rumble2 = TRUE;
                }   }
                if ((colltable[x] & 0x90) == 0x90)
                {   t |=  8; // sprites #0 and #3 collision (%1001,0,000)
#ifdef DEBUGCOLLISIONS
                    T1; printf("Sprites #0 and #3 have collided at pixel %ld!\n", x); T2;
#endif
                    if (p1sprite == 0 || p1sprite == 3)
                    {   rumble1 = TRUE;
                    }
                    if (p2sprite == 0 || p2sprite == 3)
                    {   rumble2 = TRUE;
                }   }
                if ((colltable[x] & 0x60) == 0x60)
                {   t |=  4; // sprites #1 and #2 collision (%0110,0,000)
#ifdef DEBUGCOLLISIONS
                    T1; printf("Sprites #1 and #2 have collided at pixel %ld!\n", x); T2;
#endif
                    if (p1sprite == 1 || p1sprite == 2)
                    {   rumble1 = TRUE;
                    }
                    if (p2sprite == 1 || p2sprite == 2)
                    {   rumble2 = TRUE;
                }   }
                if ((colltable[x] & 0xA0) == 0xA0)
                {   t |=  2; // sprites #1 and #3 collision (%1010,0,000)
#ifdef DEBUGCOLLISIONS
                    T1; printf("Sprites #1 and #3 have collided at pixel %ld!\n", x); T2;
#endif
                    if (p1sprite == 1 || p1sprite == 3)
                    {   rumble1 = TRUE;
                    }
                    if (p2sprite == 1 || p2sprite == 3)
                    {   rumble2 = TRUE;
                }   }
                if ((colltable[x] & 0xC0) == 0xC0)
                {   t |=  1; // sprites #2 and #3 collision (%1100,0,000)
#ifdef DEBUGCOLLISIONS
                    T1; printf("Sprites #2 and #3 have collided at pixel %ld!\n", x); T2;
#endif
                    if (p1sprite == 2 || p1sprite == 3)
                    {   rumble1 = TRUE;
                    }
                    if (p2sprite == 2 || p2sprite == 3)
                    {   rumble2 = TRUE;
            }   }   }
            pviwrite(I_SPRITECOLLIDE , (UBYTE) (memory[I_SPRITECOLLIDE ] | t));
            pviwrite(I_SPRITECOLLIDE2, (UBYTE) (memory[I_SPRITECOLLIDE2] | t));
            pviwrite(I_SPRITECOLLIDE3, (UBYTE) (memory[I_SPRITECOLLIDE3] | t));
            pviwrite(I_SPRITECOLLIDE4, (UBYTE) (memory[I_SPRITECOLLIDE4] | t));

            if (rast >= 20 && rast <= 219) // just a speed optimization
            {   t = 0;
                for (x = 0; x < I_XSIZE; x++)
                {   if ((colltable[x] & 0x18) == 0x18) // sprite #0 collision with background (%0001,1,000)
                    {    t |= 128;
                         if (p1sprite == 0) rumble1 = TRUE;
                         if (p2sprite == 0) rumble2 = TRUE;
                    }
                    if ((colltable[x] & 0x28) == 0x28) // sprite #1 collision with background (%0010,1,000)
                    {    t |= 64;
                         if (p1sprite == 1) rumble1 = TRUE;
                         if (p2sprite == 1) rumble2 = TRUE;
                    }
                    if ((colltable[x] & 0x48) == 0x48) // sprite #2 collision with background (%0100,1,000)
                    {    t |= 32;
                         if (p1sprite == 2) rumble1 = TRUE;
                         if (p2sprite == 2) rumble2 = TRUE;
                    }
                    if ((colltable[x] & 0x88) == 0x88) // sprite #3 collision with background (%1000,1,000)
                    {    t |= 16;
                         if (p1sprite == 3) rumble1 = TRUE;
                         if (p2sprite == 3) rumble2 = TRUE;
                }   }
                pviwrite(I_BGCOLLIDE , (UBYTE) (memory[I_BGCOLLIDE ] | t));
                pviwrite(I_BGCOLLIDE2, (UBYTE) (memory[I_BGCOLLIDE2] | t));
                pviwrite(I_BGCOLLIDE3, (UBYTE) (memory[I_BGCOLLIDE3] | t));
                pviwrite(I_BGCOLLIDE4, (UBYTE) (memory[I_BGCOLLIDE4] | t));
        }   }

        if (rast == 271)
        {   // this point is the start of vertical blank

            pviwrite(I_SPRITECOLLIDE , (UBYTE) (memory[I_SPRITECOLLIDE ] | 0x40)); // set VRST bit
            pviwrite(I_SPRITECOLLIDE2, (UBYTE) (memory[I_SPRITECOLLIDE2] | 0x40)); // set VRST bit
            pviwrite(I_SPRITECOLLIDE3, (UBYTE) (memory[I_SPRITECOLLIDE3] | 0x40)); // set VRST bit
            pviwrite(I_SPRITECOLLIDE4, (UBYTE) (memory[I_SPRITECOLLIDE4] | 0x40)); // set VRST bit

            psu |= PSU_S;

            if (!(psu & PSU_II) || forceinterrupt)
            {   interrupt = TRUE;

#ifdef DEBUGINTERRUPTS
                T1;
                printf("Generated vblank interrupt.\n");
                T2;
            } else
            {   T1;
                printf("Suppressed vblank interrupt.\n");
                T2;
#endif

    }   }   }

    // We are assuming that the actual diagonal (bottom-right to top-left)
    // beam retrace is done during rasters 309..311.

/*  if (rumble)
    {   if (rumble1)
        {   for (y = 0; y < IE_BOXHEIGHT; y++)
            {   ie_drawpixel(0, y, 15 - screen[0][y]);
        }   }
        if (rumble2)
        {   for (y = 0; y < IE_BOXHEIGHT; y++)
            {   ie_drawpixel(BOXWIDTH - 1, y, 15 - screen[BOXWIDTH - 1][y]);
    }   }   } */

    // Currently, frame skipping only affects warp mode when "limit
    // refresh rate" is off.

    if
    (   (!warp || !limitrefresh)
     && (frames % frameskip == 0)
    )
    {   updatescreen();
    }

    frames++;

    inframe = FALSE;
    if (crippled)
    {   uncripple();
    }

    if (recmode == RECMODE_PLAY && offset >= recsize)
    {   if (loop)
        {   macro_restartplayback();
        } else
        {   macro_stop();
    }   }

    if (runtoframe)
    {   runtoframe = FALSE;
        T1;
        printf("Reached next frame (%ld).\n\n", (long int) frames);
        T2;
        pause();
}   }

MODULE __inline void breakrastline(void)
{   if (runtorastline)
    {   runtorastline = FALSE;
        T1;
        printf("Reached next raster line (%ld).\n\n", (long int) rast);
        T2;
        runto();
    }
    if (rast == rastn)
    {   rastn = -1;
        T1;
        printf("Reached raster line %ld.\n\n", (long int) rast);
        T2;
        runto();
}   }

MODULE __inline void drawgrid(void)
{   PERSIST int gridline[220] =                                                         // Interton/Elektor screen is vertically as follows:
    {   20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, //   0.. 19 top (visible) margin
         0,  0,                                                                         //  20.. 21 thin
                 1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1, //  22.. 39 thick
         2,  2,                                                                         //  40.. 41 thin
                 3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3, //  42.. 59 thick
         4,  4,                                                                         //  60.. 61 thin
                 5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5, //  62.. 79 thick
         6,  6,                                                                         //  80.. 81 thin
                 7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7, //  82.. 99 thick
         8,  8,                                                                         // 100..101 thin
                 9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9, // 102..119 thick
        10, 10,                                                                         // 120..121 thin
                11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, // 122..139 thick
        12, 12,                                                                         // 140..141 thin
                13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, // 142..159 thick
        14, 14,                                                                         // 160..161 thin
                15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, // 162..179 thick
        16, 16,                                                                         // 180..181 thin
                17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, // 182..199 thick
        18, 18,                                                                         // 200..201 thin
                19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19  // 202..219 thick
                                                                                        // 220..271 bottom visible margin
                                                                                        // 272..311 bottom invisible margin
    };
    PERSIST int i, j, k,
                m,
                w;

    // assert(rast < BOXHEIGHT);

#ifdef DEBUGSPRITES
//    T1;
//    printf("Drawing grid for rastline %ld.\n", rast);
//    T2;
#endif

    if (flagline)
    {   if (pviread(I_BGCOLOUR) & 8) // background/grid enable
        {   ie_drawline(15 -  (pviread(I_BGCOLOUR) &  7)    );
            fgc =       15 - ((pviread(I_BGCOLOUR) >> 4) & 7);
        } else
        {   ie_drawline(15); // dark black (black)
            return;
    }   }
    else
    {   if (pviread(I_BGCOLOUR) & 8) // background/grid enable
        {   ie_drawline(7 -  (pviread(I_BGCOLOUR) &  7)    );
            fgc =       7 - ((pviread(I_BGCOLOUR) >> 4) & 7);
        } else
        {   ie_drawline(7); // black
            return;
    }   }

#ifndef NOGRID
    if (rast < 20 || rast > 219)
#endif
        return;

    if
    (   (pviread(I_NOISE) & 0x20)
     && (machine == INTERTON || memmap == MEMMAP_F)
    )
    {   if (fgc >= 8)
        {   fgc = 15 - fgc; // 0..7
        } else
        {   fgc =  7 - fgc; // 0..7
    }   }

    i = gridline[rast];              //  i       is     0..19
    k = pviread(I_HORIZ1 + (i >> 2)); // (i >> 2) yields 0.. 4
    switch (k & 0xC0)
    {
    case  0x40:
        w = 2;
    acase 0xC0:
        w = 4;
        adefault: // $00, $80
        w = 1;
    break;
    }

    switch (i & 3)                   // (i &  3) yields 0.. 3
    {
    case 0:
        if (k & 1)
        {   w = 8;
        }
    acase 1:
        if (((rast - 20) % 40) <= 10)
        {   if (k & 2)
            {   w = 8;
        }   }
        else
        {   if (k & 4)
            {   w = 8;
        }   }
    acase 2:
        if (k & 8)
        {   w = 8;
        }
    acase 3:
        if (((rast - 20) % 40) <= 30)
        {   if (k & 16)
            {   w = 8;
        }   }
        else
        {   if (k & 32)
            {   w = 8;
        }   }
    break;
    }

    // j goes   0,  1,  2,  3,  4,  5,  6,  7,   8,   9,  10,  11,  12,  13,  14,  15
    // m goes 128, 64, 32, 16,  8,  4,  2,  1, 128,  64,  32,  16,   8,   4,   2,   1
    // x goes  32, 42, 50, 58, 66, 74, 82, 90,  98, 106, 114, 122, 130, 138, 146, 154
    for
    (   x = 32, j = 0, m = 128; // was x = 31 until V4.62
        j < 16;
        j++, x += 8, m >>= 1
    )
    {   if (pviread(I_VERTGRID + (i * 2) + (j >> 3)) & m)
        {   for (xx = x; xx < x + w; xx++)
            {   colltable[xx] |= 8;
                ie_drawpixel
                (   xx - HIDDEN_X,
                    rast,
                    fgc
                );
        }   }
        if (j == 7)
        {   m = 256;
}   }   }

/* Gridlines that are the same colour as the background (ie. invisible)
   don't register collisions. This seems to be the correct behaviour.

     -0-  -1-  -2-  -3-  -4-  -5-  -6-  -7-  -8-  -9-  -A-  -B-  -C-  -D-  -E-  -F-
abc  ###  ..#  ###  ###  #.#  ###  ###  ###  ###  ###  ###  #..  ###  ..#  ###  ###
l d  # #  . #  ..#  . #  # #  # .  # .  . #  # #  # #  # #  # .  # .  . #  # .  # .
kme  #.#  ..#  ###  ###  ###  ###  ###  ..#  ###  ###  ###  ###  #..  ###  ###  ###
j f  # #  . #  #..  . #  . #  . #  # #  . #  # #  . #  # #  # #  # .  # #  # .  # .
ihg  ###  ..#  ###  ###  ..#  ###  ###  ..#  ###  ###  #.#  ###  ###  ###  ###  #..

The Interton/Elektor handles $A..$F values by drawing nothing.
The capital $B looks the same as 8
and capital $D looks the same as 0
so we would use lowercase "b" and "d". */

MODULE void drawdigit(int x, int y, int d, int line)
{   // bit $1000: 'm' segment
    // bit $0800: 'l' segment
    // bit $0400: 'k' segment
    // bit $0200: 'j' segment
    // bit $0100: 'i' segment
    // bit $0080: 'h' segment
    // bit $0040: 'g' segment
    // bit $0020: 'f' segment
    // bit $0010: 'e' segment
    // bit $0008: 'd' segment
    // bit $0004: 'c' segment
    // bit $0002: 'b' segment
    // bit $0001: 'a' segment
    PERSIST const UWORD digit_to_segment[10] = // [16] for hex support
    {   0x0fff, // 0
        0x007c, // 1
        0x17df, // 2
        0x15ff, // 3
        0x1c7d, // 4
        0x1df7, // 5
        0x1ff7, // 6
        0x007f, // 7
        0x1fff, // 8
        0x1dff  // 9
/*    , 0x1f7f, // A
        0x1ff1, // B
        0x0fc7, // C
        0x17fc, // D
        0x1fd7, // E
        0x1f17  // F
*/
    };
    int j;

    // assert(d >=  0);
    // assert(d <= 15);
    if (d >= 10)
    {   return;
    }

    for (j = 0; j < 12 + 1; j++)
    {   if
        (   digit_to_segment[d] & (1 << (led[line][j] - 'a'))
        )
        {   colltable[x + j - HIDDEN_X] |= 8;
            ie_drawpixel(x + j, y + line, digitcolour);
}   }   }

MODULE __inline void drawsprites(void)
{   PERSIST int   i;
    PERSIST int   thesize[4];
    PERSIST UBYTE spriteimage[4];

    // assert(rast < 272);

    /* delay is used to stretch out the sprite horizontally (it counts up
       until it reaches thesize) */

    for (whichsprite = 0; whichsprite < 4; whichsprite++)
    {

#ifdef DEBUGSPRITES
        T1;
        printf("Rastline %ld, sprite %ld, state %ld.\n", rast, whichsprite, sprite[whichsprite].state);
        T2;
#endif

        /* It is probable that mid-sprite changes to SIZE are ineffective
           (this behaviour helps HUNTING, SHOOTGAL). */

        switch (sprite[whichsprite].state)
        {
        case 0: // top of sprite not yet reached by the beam
            if (rast == sprite[whichsprite].y)
            {   sprite[whichsprite].delay = 0;
                sprite[whichsprite].state = 1; // starts drawing immediately (this rastline)?
                if (whichsprite == 0)
                {   thesize[0] = 1 << ( pviread(I_SIZES)       & 3);
                } elif (whichsprite == 1)
                {   thesize[1] = 1 << ((pviread(I_SIZES) >> 2) & 3);
                } elif (whichsprite == 2)
                {   thesize[2] = 1 << ((pviread(I_SIZES) >> 4) & 3);
                } else
                {   // assert(whichsprite == 3);
                    thesize[3] = 1 << ((pviread(I_SIZES) >> 6) & 3);
            }   }
            else
            {   break;
            }
        case 1: // drawing the top row
        case 2: // drawing the 2nd row
        case 3: // etc.
        case 4:
        case 5:
        case 6:
        case 7:
        case 8:
        case 9:
        case 10:
            if (sprite[whichsprite].delay == 0)
            {   spriteimage[whichsprite] = pviread(ie_spritedata[whichsprite]
                                         + sprite[whichsprite].state
                                         - 1);
                if (whichsprite == 0)
                {   sprite[0].colour = ((pviread(I_SPR01COLOURS) >> 3) & 7);
                } elif (whichsprite == 1)
                {   sprite[1].colour = ( pviread(I_SPR01COLOURS)       & 7);
                } elif (whichsprite == 2)
                {   sprite[2].colour = ((pviread(I_SPR23COLOURS) >> 3) & 7);
                } else
                {   // assert(whichsprite == 3);
                    sprite[3].colour = ( pviread(I_SPR23COLOURS)       & 7);
            }   }
            xx = pviread(ie_spritedata[whichsprite] + 10) + 1; // SPRITEnAX
            if (xx < I_XSIZE)
            {
#ifdef DEBUGSPRITES
                T1;
                printf
                (   "Drawing row %ld of sprite #%ld (1st) at %ld,%ld (image data is $%lX).\n",
                    sprite[whichsprite].state,
                    whichsprite,
                    xx, // it is given in real Interton/Elektor coords, not ie_drawpixel()-adjusted
                    rast,
                    spriteimage[whichsprite]
                );
                T2;
#endif
#ifdef NOSPRITES
                goto DONE1;
#endif
#ifdef SELECTIVESPRITES
                if (!enabled[whichsprite]) goto DONE1;
#endif
                for (i = 0; i < 8; i++) // the 8 columns of the sprite
                {   if (spriteimage[whichsprite] & (128 >> i))
                    {   for (x = 0; x < thesize[whichsprite]; x++, xx++)
                        {   if (xx >= I_XSIZE)
                            {   goto DONE1;
                            }
                            colltable[xx] |= (0x10 << whichsprite) | sprite[whichsprite].colour;
                            if
                            (   rast <  BOXHEIGHT
                             && xx   >= HIDDEN_X
                             && xx   <  HIDDEN_X + BOXWIDTH
                            )
                            {   ie_drawpixel
                                (   xx - HIDDEN_X,
                                    rast,
                                    colltable[xx] & 7
                                );
                    }   }   }
                    else
                    {   xx += thesize[whichsprite];
            }   }   }

DONE1:
            sprite[whichsprite].delay++;
            if (sprite[whichsprite].delay >= thesize[whichsprite])
            {   sprite[whichsprite].delay = 0;
                sprite[whichsprite].state++;
                if (sprite[whichsprite].state == 11)
                {   /* Old sprite has been drawn, now we reread SPRITEnBY.
                   This must happen now (before running the CPU), not later!

"Unlike SPRITEnAX and SPRITEnAY, SPRITEnBX and SPRITEnBY may
be changed duiring the scan [(frame)] and such changes will be effected
on the current scan [(frame)]. SPRITEnBX is sampled during each HRST
[(horizontal reset)]. Thus, if SPRITEnBX is changed during the
display of a duplicate, a portion of the object will be
displaced horizontally. For proper operation, SPRITEnBX should
be changed only *after* the 'Object Video Completion' status
bit [in BGCOLLIDE] indicates completion of the object.
SPRITEnBY is sampled just prior to displaying the last line of
the object. To effect a change in the vertical offset to the
next duplicate, SPRITEnBY must be changed *before* the 'Object
Video Completion' status bit [in BGCOLLIDE] indicates
completion of the object." - PVI manual, p. 5. */

                    sprite[whichsprite].y = pviread(ie_spritedata[whichsprite] + 13); // SPRITEnBY
                    sprite[whichsprite].y++; // overflow OK!

                    pviwrite(I_BGCOLLIDE , (UBYTE) (memory[I_BGCOLLIDE ] | (8 >> whichsprite)));
                    pviwrite(I_BGCOLLIDE2, (UBYTE) (memory[I_BGCOLLIDE2] | (8 >> whichsprite)));
                    pviwrite(I_BGCOLLIDE3, (UBYTE) (memory[I_BGCOLLIDE3] | (8 >> whichsprite)));
                    pviwrite(I_BGCOLLIDE4, (UBYTE) (memory[I_BGCOLLIDE4] | (8 >> whichsprite)));
                    if (!(psu & PSU_II) || forceinterrupt)
                    {   interrupt = TRUE;

#ifdef DEBUGINTERRUPTS
                        T1;
                        printf("Generated original sprite #%ld display complete interrupt.\n", whichsprite);
                    T2;
                    } else
                    {   T1;
                        printf("Suppressed original sprite #%ld display complete interrupt.\n", whichsprite);
                        T2;
#endif

            }   }   }
        break;
        case 11:
            if (sprite[whichsprite].y <= 252) // $00..$FC
            {   sprite[whichsprite].delay = 0;
                sprite[whichsprite].state = 12;
            } else // $FD..$FF
            {   break;
            }
        case 12:
            // count down the vertical gap
            if (sprite[whichsprite].y == 0 && rast <= fudge)
            {   sprite[whichsprite].state = 13; // starts drawing immediately (this rastline)
                if (whichsprite == 0)
                {   thesize[0] = 1 << ( pviread(I_SIZES)       & 3);
                } elif (whichsprite == 1)
                {   thesize[1] = 1 << ((pviread(I_SIZES) >> 2) & 3);
                } elif (whichsprite == 2)
                {   thesize[2] = 1 << ((pviread(I_SIZES) >> 4) & 3);
                } else
                {   // assert(whichsprite == 3);
                    thesize[3] = 1 << ((pviread(I_SIZES) >> 6) & 3);
            }   }
            else
            {   // this must be done even when (sprite[whichsprite].y == 0 && rast > fudge)!
                // (Otherwise breaks Newton)
                sprite[whichsprite].y--;
                break;
            }
        case 13:
        case 14:
        case 15:
        case 16:
        case 17:
        case 18:
        case 19:
        case 20:
        case 21:
        case 22:
            if (sprite[whichsprite].delay == 0)
            {   spriteimage[whichsprite] = pviread(ie_spritedata[whichsprite]
                     + sprite[whichsprite].state
                     - 13);
                if (whichsprite == 0)
                {   sprite[0].colour = ((pviread(I_SPR01COLOURS) >> 3) & 7);
                } elif (whichsprite == 1)
                {   sprite[1].colour = ( pviread(I_SPR01COLOURS)       & 7);
                } elif (whichsprite == 2)
                {   sprite[2].colour = ((pviread(I_SPR23COLOURS) >> 3) & 7);
                } else
                {   // assert(whichsprite == 3);
                    sprite[3].colour = ( pviread(I_SPR23COLOURS)       & 7);
            }   }
            xx = pviread(ie_spritedata[whichsprite] + 11) + 1; // SPRITEnBX
            if (xx < I_XSIZE)
            {
#ifdef DEBUGSPRITES
                T1;
                printf
                (   "Drawing row %ld of sprite #%ld (2nd) at %ld,%ld (image data is $%lX).\n",
                    sprite[whichsprite].state - 12,
                    whichsprite,
                    xx, // it is given in real Interton/Elektor coords, not ie_drawpixel()-adjusted
                    rast,
                    spriteimage[whichsprite]
                );
                T2;
#endif
#ifdef NOSPRITES
                goto DONE2;
#endif
#ifdef SELECTIVESPRITES
                if (!enabled[whichsprite]) goto DONE2;
#endif
                for (i = 0; i < 8; i++) // the 8 columns of the sprite
                {   if (spriteimage[whichsprite] & (128 >> i))
                    {   for (x = 0; x < thesize[whichsprite]; x++, xx++)
                        {   if (xx >= I_XSIZE)
                            {   goto DONE2;
                            }
                            colltable[xx] |= (0x10 << whichsprite) | sprite[whichsprite].colour;
                            if
                            (   rast <  BOXHEIGHT
                             && xx   >= HIDDEN_X
                             && xx   <  HIDDEN_X + BOXWIDTH
                            )
                            {   ie_drawpixel
                                (   xx - HIDDEN_X,
                                    rast,
                                    colltable[xx] & 7
                                );
                    }   }   }
                    else
                    {   xx += thesize[whichsprite];
            }   }   }

DONE2:
            sprite[whichsprite].delay++;
            if (sprite[whichsprite].delay >= thesize[whichsprite])
            {   sprite[whichsprite].delay = 0;
                sprite[whichsprite].state++;

                if (sprite[whichsprite].state == 23)
                {   // old sprite has been drawn, now we reread SPRITEnBY
                    // this must happen now (before running the CPU), not later!
                    sprite[whichsprite].y = pviread(ie_spritedata[whichsprite] + 13); // SPRITEnBY
                    sprite[whichsprite].y++; // overflow OK!

                    pviwrite(I_BGCOLLIDE , (UBYTE) (memory[I_BGCOLLIDE ] | (8 >> whichsprite)));
                    pviwrite(I_BGCOLLIDE2, (UBYTE) (memory[I_BGCOLLIDE2] | (8 >> whichsprite)));
                    pviwrite(I_BGCOLLIDE3, (UBYTE) (memory[I_BGCOLLIDE3] | (8 >> whichsprite)));
                    pviwrite(I_BGCOLLIDE4, (UBYTE) (memory[I_BGCOLLIDE4] | (8 >> whichsprite)));
                    if (!(psu & PSU_II) || forceinterrupt)
                    {   interrupt = TRUE;

#ifdef DEBUGINTERRUPTS
                        T1;
                        printf("Generated duplicate sprite #%ld display complete interrupt.\n", whichsprite);
                        T2;
                    } else
                    {   T1;
                        printf("Suppressed duplicate sprite #%ld display complete interrupt.\n", whichsprite);
                        T2;
#endif

                    }
                    sprite[whichsprite].state = 11;
            }   }
        adefault:
        break;
}   }   }

EXPORT void pviwrite(UWORD address, UBYTE data)
{   /* Writes to memory from the PVI always go through this
       routine (so that watchpoints work properly). */

    if (address == (UWORD) wp && data != memory[wp])
    {   getfriendly(wp);
        T1;
        printf("PVI is writing $%X to %s at raster %ld!\n\n", data, friendly, (long int) rast);
        T2;
        runto();
    }

    memory[address] = data;
}

// We could do uviread() too.
// Networking, paddles, macro code isn't using pviread()/pviwrite().
MODULE __inline UBYTE pviread(int address)
{   /* if
    (   address != I_BGCOLOUR
     && address != I_NOISE
     && (address < 0x1F80 || address > 0x1FAC)
    )
    {   getfriendly(address);
        T1;
        printf("PVI is reading $%X from %s at frame %ld, raster %ld!\n\n", memory[address], friendly, frames, (long int) rast);
        T2;
    } */

    return(memory[address]);
}

MODULE void ie_playerinput(int source, int dest)
{   // dest is which side   (0 or 1) you want to set the registers of.
    // source is which side (0 or 1) you want to use to do it.

    if (recmode == RECMODE_PLAY)
    {   // I_P1PADDLE is $1FCC, I_P2PADDLE is $1FCD
        t = IOBuffer[offset++];
        pviwrite((UWORD) (I_P1PADDLE     +  dest     ),            t                      );
        t = IOBuffer[offset++];
        pviwrite((UWORD) (I_P1LEFTKEYS   + (dest * 4)), (UBYTE) ( (t & 0xF0)       | 0x0F)); // bits 7..4 -> bits  7..4 (masking is indeed necessary)
        pviwrite((UWORD) (I_P1MIDDLEKEYS + (dest * 4)), (UBYTE) (((t & 0x0F) << 4) | 0x0F)); // bits 3..0 -> bits  7..4
        t = IOBuffer[offset++];
        pviwrite((UWORD) (I_P1RIGHTKEYS  + (dest * 4)),            t                      );
    } else
    {   // assert(recmode == RECMODE_NORMAL || recmode == RECMODE_RECORD);

        if
        (   (dest == 0 && connected == NET_CLIENT)
         || (dest == 1 && connected == NET_SERVER)
        )
        {   return;
        }

        joystate = ReadJoystick((UWORD) (1 - source));

        memory[I_P1LEFTKEYS   + (dest * 4)] =
        memory[I_P1MIDDLEKEYS + (dest * 4)] =
        memory[I_P1RIGHTKEYS  + (dest * 4)] = 0x0F;
        if (autofire)
        {   if
            (   (!requirebutton || (joystate & JOYFIRE1) || KeyDown(keypads[source][0]))
             && ((frames % totalframes) < downframes)
            )
            {   // 1st firebutton
                memory[keytable[key1].ie_address + (dest * 4)] |= keytable[key1].ie_mask;
        }   }
        else
        {   if ((joystate & JOYFIRE1) || KeyDown(keypads[source][0]))
            {   // 1st firebutton
                memory[keytable[key1].ie_address + (dest * 4)] |= keytable[key1].ie_mask;
        }   }
        if
        (   (joystate & JOYFIRE2)
         || (KeyDown(keypads[source][21]))
        )
        {   // 2nd firebutton
            memory[keytable[key2].ie_address + (dest * 4)] |= keytable[key2].ie_mask;
        }
        if
        (   (joystate & JOYFIRE3)
         || (KeyDown(keypads[source][22]))
        )
        {   // 3rd firebutton
            memory[keytable[key3].ie_address + (dest * 4)] |= keytable[key3].ie_mask;
        }
        if
        (   (joystate & JOYFIRE4)
         || (KeyDown(keypads[source][23]))
        )
        {   // 4th firebutton
            memory[keytable[key4].ie_address + (dest * 4)] |= keytable[key4].ie_mask;
        }

        // left column
        t = memory[I_P1LEFTKEYS + (dest * 4)];
        t |= (0x80 * KeyDown(keypads[source][ 1])); // "1"  key (Interton) or "RCAS"  key (Elektor)
        t |= (0x40 * KeyDown(keypads[source][ 4])); // "4"  key (Interton) or "BP1/2" key (Elektor)
        t |= (0x20 * KeyDown(keypads[source][ 7])); // "7"  key (Interton) or "PC"    key (Elektor)
        t |= (0x10 * KeyDown(keypads[source][10])); // "Cl" key (Interton) or "-"     key (Elektor)
        pviwrite((UWORD) (I_P1LEFTKEYS   + (dest * 4)), t);

        // middle column
        t = memory[I_P1MIDDLEKEYS + (dest * 4)];
        t |= (0x80 * KeyDown(keypads[source][ 2])); // "2"  key (Interton) or "WCAS" key (Elektor)
        t |= (0x40 * KeyDown(keypads[source][ 5])); // "5"  key (Interton) or "REG"  key (Elektor)
        t |= (0x20 * KeyDown(keypads[source][ 8])); // "8"  key (Interton) or "MEM"  key (Elektor)
        t |= (0x10 * KeyDown(keypads[source][11])); // "0"  key (Interton) or "+"    key (Elektor)
        pviwrite((UWORD) (I_P1MIDDLEKEYS + (dest * 4)), t);

        // right column
        t = memory[I_P1RIGHTKEYS + (dest * 4)];
        t |= (0x80 * KeyDown(keypads[source][ 3])); // "3"  key (Interton) or "C" key (Elektor)
        t |= (0x40 * KeyDown(keypads[source][ 6])); // "6"  key (Interton) or "8" key (Elektor)
        t |= (0x20 * KeyDown(keypads[source][ 9])); // "9"  key (Interton) or "4" key (Elektor)
        t |= (0x10 * KeyDown(keypads[source][12])); // "En" key (Interton) or "0" key (Elektor)
        pviwrite((UWORD) (I_P1RIGHTKEYS  + (dest * 4)), t);

        dopaddle(source, dest);

        if (psu & PSU_F) // paddle interpolation bit
        {   pviwrite((UWORD) (I_P1PADDLE + dest), (UBYTE) ay[dest]);
        } else
        {   pviwrite((UWORD) (I_P1PADDLE + dest), (UBYTE) ax[dest]);
    }   }

    if (psu & PSU_F) // paddle interpolation bit
    {   sy[dest] = (UBYTE) ay[dest];
    } else
    {   sx[dest] = (UBYTE) ax[dest];
}   }

MODULE __inline void ie_emuinput(void)
{   UBYTE therecv[3],
          thesend[3];

    readkybd();

    // must always do ie_playerinput(foo, 0) then ie_playerinput(foo, 1).
    if (swapped)
    {   ie_playerinput(1, 0);
        ie_playerinput(0, 1);
    } else
    {   ie_playerinput(0, 0);
        ie_playerinput(1, 1);
    }

    if (recmode != RECMODE_PLAY)
    {

#ifdef AMIGA
        if (absolute)
        {   joystate = 0;
        } else
#endif

            joystate = ReadJoystick(1) | ReadJoystick(0);

        if (KeyDown(console[0]) || (joystate & JOYSTART)) // START /START
        {   t = 0x4F;
        } else
        {   t = 0x0F;
        }
        if (KeyDown(console[1]) || (joystate & JOYA    )) // SELECT/UC
        {   t |= 0x80;
        }
        if (machine == ELEKTOR)
        {   if (KeyDown(console[2]) || (joystate & JOYB)) // -/LC
            {   t |= 0x20;
            }
            if (KeyDown(console[3])                     ) // -/RESET (soft)
            {   t |= 0x10;
        }   }

        pviwrite(I_CONSOLE, t);

        domouse();
    }

    if (connected == NET_SERVER)
    {   // we are left player

        thesend[0] = ((memory[I_P1LEFTKEYS  ] & 0x0F) <<  4) // bits 3..0 -> bits  7..4 (masking not really needed)
                   |  (memory[I_P1MIDDLEKEYS] & 0x0F);       // bits 3..0 -> bits  3..0 (masking not really needed)
        thesend[1] =  (memory[I_P1RIGHTKEYS ] & 0x0F)        // bits 3..0 -> bits  3..0 (masking not really needed)
                   |  (memory[I_CONSOLE     ] & 0xF0);       // bits 7..4 -> bits  7..4 (masking not really needed)
        thesend[2] =   memory[I_P1PADDLE];                   // bits 7..0 -> bits  7..0

        if (NetExchange((char *) &therecv, (const char *) &thesend, 3))
        {   pviwrite(I_P2LEFTKEYS  , (UBYTE) (((therecv[0] & 0xF0) >> 4) | 0x0F)); // bits 7..4 -> bits 3..0
            pviwrite(I_P2MIDDLEKEYS, (UBYTE) (( therecv[0] & 0x0F      ) | 0x0F)); // bits 3..0 -> bits 3..0
            pviwrite(I_P2RIGHTKEYS , (UBYTE) (( therecv[1] & 0x0F      ) | 0x0F)); // bits 3..0 -> bits 3..0 (masking not really needed)
            pviwrite(I_P2PADDLE    ,            therecv[2]                      ); // bits 7..0 -> bits 7..0
            pviwrite(I_CONSOLE     , (UBYTE) (( therecv[1] & 0xF0      ) | memory[I_CONSOLE]));
    }   }
    elif (connected == NET_CLIENT)
    {   // we are right player

        thesend[0] = ((memory[I_P2LEFTKEYS  ] & 0x0F) <<  4) // bits 3..0 -> bits  7..4 (masking not really needed)
                   |  (memory[I_P2MIDDLEKEYS] & 0x0F);       // bits 3..0 -> bits  3..0 (masking not really needed)
        thesend[1] =  (memory[I_P2RIGHTKEYS ] & 0x0F)        // bits 3..0 -> bits  3..0 (masking not really needed)
                   |  (memory[I_CONSOLE     ] & 0xF0);       // bits 7..4 -> bits  7..4 (masking not really needed)
        thesend[2] =   memory[I_P2PADDLE];                   // bits 7..0 -> bits  7..0

        if (NetExchange((char *) &therecv, (const char *) &thesend, 3))
        {   pviwrite(I_P1LEFTKEYS  , (UBYTE) (((therecv[0] & 0xF0) >> 4) | 0x0F)); // bits 7..4 -> bits 3..0
            pviwrite(I_P1MIDDLEKEYS, (UBYTE) (( therecv[0] & 0x0F      ) | 0x0F)); // bits 3..0 -> bits 3..0
            pviwrite(I_P1RIGHTKEYS , (UBYTE) (( therecv[1] & 0x0F      ) | 0x0F)); // bits 3..0 -> bits 3..0 (masking not really needed)
            pviwrite(I_P1PADDLE    ,            therecv[2]                      ); // bits 7..0 -> bits 7..0
            pviwrite(I_CONSOLE     , (UBYTE) (( therecv[1] & 0xF0      ) | memory[I_CONSOLE]));
    }   }

    if (recmode == RECMODE_PLAY)
    {   pviwrite(I_CONSOLE, IOBuffer[offset++]);
    } elif (recmode == RECMODE_RECORD)
    {   // assert(MacroHandle);

        OutputBuffer[0] =   memory[I_P1PADDLE    ];
        OutputBuffer[1] =  (memory[I_P1LEFTKEYS  ] & 0xF0)         // bits 7..4 -> bits  7..4 (masking not really needed)
                        | ((memory[I_P1MIDDLEKEYS] & 0xF0) >>  4); // bits 7..4 -> bits  3..0 (masking not really needed)
        OutputBuffer[2] =   memory[I_P1RIGHTKEYS ];                // 4 bits are wasted
        OutputBuffer[3] =   memory[I_P2PADDLE    ];
        OutputBuffer[4] =  (memory[I_P2LEFTKEYS  ] & 0xF0)         // bits 7..4 -> bits  7..4 (masking not really needed)
                        | ((memory[I_P2MIDDLEKEYS] & 0xF0) >>  4); // bits 7..4 -> bits  3..0 (masking not really needed)
        OutputBuffer[5] =   memory[I_P2RIGHTKEYS ];                // 4 bits are wasted
        OutputBuffer[6] =   memory[I_CONSOLE     ];                // 6 (Interton) or 4 (Elektor) bits are wasted

        DISCARD fwrite(OutputBuffer, 7, 1, MacroHandle);
        /* fwrite() should really check return code
        7 bytes per frame * 50 frames per second = 350 bytes/sec.
        Plus, for Elektor, 1 byte every time the E_RANDOMn hardware
        registers are read. */
}   }

MODULE __inline void ie_drawpixel(int x, int y, int colour)
{   // assert(x >= 0);
    // assert(x < BOXWIDTH);
    // assert(y >= 0);
    // assert(y < BOXHEIGHT);
    // assert(colour >= 0);
    // assert(colour <= 15);

#ifdef DEBUGSPRITES
    if (x < 0)
    {   T1;
        printf("X of %ld < 0!\n", x);
        T2;
        return;
    }
    if (x >= BOXWIDTH)
    {   T1;
        printf("X of %ld >= BOXWIDTH of %ld!\n", x, BOXWIDTH);
        T2;
        return;
    }
    if (y < 0)
    {   T1;
        printf("Y of %ld < 0!\n", y);
        T2;
        return;
    }
    if (y >= BOXHEIGHT)
    {   T1;
        printf("Y of %ld >= BOXHEIGHT of %ld!\n", y, BOXHEIGHT);
        T2;
        return;
    }
    if (colour > 15)
    {   T1;
        printf("Illegal colour!\n");
        T2;
        return;
    }
#endif

    if (screen[x][y] == colour)
    {   return;
    }
    screen[x][y] = colour;

#ifdef WIN32
    display[x + (y * BOXWIDTH)] = pens[colour];
#endif
#ifdef AMIGA
    if (fullscreen) // 2*1
    {   *(wordptr[y] + x) = wordpens[colour];
    } else
    {   switch(size)
        {
        case 1:
            if (wide == 1) // 1*1
            {   *(byteptr[y] + x) = pens[colour]; // 1,1
            } else
            {   // assert(wide == 2); // 2*1

                *(wordptr[y] + x) = wordpens[colour];
            }
        acase 2:
            py = y << 1; // py = y * 2;

            if (wide == 1) // 2*2
            {   *(wordptr[py    ] + x) =
                *(wordptr[py + 1] + x) = wordpens[colour];
            } else
            {   // assert(wide == 2); // 4*2

                *(longptr[py    ] + x) =
                *(longptr[py + 1] + x) = longpens[colour];
            }
        adefault:
            // assert(size == 3);

            px = x * 3;
            py = y * 3;

            if (wide == 1) // 3*3
            {   *(byteptr[py    ] + px    ) =
                *(byteptr[py + 1] + px    ) =
                *(byteptr[py + 2] + px    ) =
                *(byteptr[py    ] + px + 1) =
                *(byteptr[py + 1] + px + 1) =
                *(byteptr[py + 2] + px + 1) =
                *(byteptr[py    ] + px + 2) =
                *(byteptr[py + 1] + px + 2) =
                *(byteptr[py + 2] + px + 2) = pens[colour];
            } else
            {   // assert(wide == 2); // 6*3

                *(wordptr[py    ] + px    ) =
                *(wordptr[py + 1] + px    ) =
                *(wordptr[py + 2] + px    ) =
                *(wordptr[py    ] + px + 1) =
                *(wordptr[py + 1] + px + 1) =
                *(wordptr[py + 2] + px + 1) =
                *(wordptr[py    ] + px + 2) =
                *(wordptr[py + 1] + px + 2) =
                *(wordptr[py + 2] + px + 2) = wordpens[colour];
            }
        break;
    }   }

#endif

}

EXPORT UBYTE __inline loadrand(void)
{   // assert(recmode == RECMODE_PLAY);
    // assert(machine == ELEKTOR);
    return(IOBuffer[offset++]);
}
EXPORT void __inline saverand(UBYTE value)
{   // assert(recmode == RECMODE_RECORD);
    // assert(machine == ELEKTOR);
    OutputBuffer[0] = value;
    DISCARD fwrite(OutputBuffer, 1, 1, MacroHandle); // should really check return code
}

MODULE __inline void ie_drawline(int colour)
{
#ifdef WIN32
    PERSIST int end; // PERSISTent for speed
#endif

    // assert(rast >= 0);
    // assert(rast < BOXHEIGHT);
    // assert(colour >= 0);
    // assert(colour <= 15);

    if
    (   (machine == INTERTON || memmap == MEMMAP_F)
     && (pviread(I_NOISE) & 0x20)
    )
    {   if (colour >= 8)
        {   colour = 15 - colour; // 0..7
        } else
        {   colour =  7 - colour; // 0..7
    }   }

    for (x = 0; x < BOXWIDTH; x++)
    {   screen[x][rast] = colour;
    }

#ifdef WIN32
    end = (rast + 1) * BOXWIDTH;
    for (x = rast * BOXWIDTH; x < end; x++)
    {   display[x] = pens[colour];
    }
#endif
#ifdef AMIGA
    if (fullscreen)
    {   y = rast;
    } else y = rast * size;

    for (x = 0; x < divided; x++)
    {   *(longptr[y] + x) = longpens[colour];
    }
    if (!fullscreen && size >= 2)
    {   y++;
        for (x = 0; x < divided; x++)
        {   *(longptr[y] + x) = longpens[colour];
        }
        if (size >= 3)
        {   y++;
            for (x = 0; x < divided; x++)
            {   *(longptr[y] + x) = longpens[colour];
    }   }   }

    /* assert(!remainder);
    if (remainder)
    {   for (x = 0; x < remainder; x++)
        {   *(byteptr[rast] + (divided << 2) + x) = pens[colour];
    }   } */
#endif
}

