// uvi.c: Signetics 2637 UVI emulation module (platform-independent)

#ifdef AMIGA
    #include "lint.h"
    #include "amiga.h"

    #include <proto/locale.h> // GetCatalogStr()
#endif
#ifdef WIN32
    #include "ibm.h"
    #define EXEC_TYPES_H
#endif

#include "aa.h"
#define CATCOMP_NUMBERS
#define CATCOMP_CODE
#define CATCOMP_BLOCK
#include "aa_strings.h"
#include "uvi.h"

#include <stdio.h>
#include <string.h>

#define CPL        52 // CYCLESPERLINE is a reserved macro under Windows
// Not necessarily the same number of CPU cycles per raster line as
// for the PVI.
#define INPUTLINE 207
#define LATCHLINE 225
#define SENSELINE 240

#define FLIPS       2

IMPORT ULONG                 analog,
                             autofire,
                             collisions,
                             demultiplex,
                             downframes,
                             flagline,
                             joy1,
                             onframes,
                             paused,
                             swapped,
                             totalframes,
                             totalsprframes,
                             warp,
                             elapsed,
                             frames;
IMPORT UBYTE*                IOBuffer;
IMPORT FLAG                  absolute;
IMPORT FILE*                 MacroHandle;
IMPORT TEXT                  friendly[2 + 13 + 1 + 1];
IMPORT int                   connected,
                             consoleopen,
                             crippled,
                             frameskip,
                             inframe,
                             key1,
                             key2,
                             key3,
                             key4,
                             limitrefresh,
                             loop,
                             memmap,
                             offset,
                             overcalc, // must be signed!
                             p1sprite,
                             p2sprite,
                             rast,
                             rastn,
                             recmode,
                             recsize,
                             requirebutton,
                             runtorastline,
                             runtodma,
                             runtoframe,
                             sensitivity,
                             slice,
                             trace,
                             wp;
IMPORT UBYTE                 memory[32768],
                             screen[BOXWIDTH][BOXHEIGHT],
                             psu,
                             sx[2], sy[2];
IMPORT UWORD                 console[4],
                             joystate,
                             keypads[2][NUMKEYS],
                             page,
                             iar;
IMPORT SLONG                 ax[2], // analog paddle X-coords
                             ay[2]; // analog paddle Y-coords
IMPORT struct KeyTableStruct keytable[16];

#ifdef WIN32
    IMPORT ULONG*          display;
    IMPORT ULONG           pens[16];
    IMPORT int             CatalogPtr;
#endif
#ifdef AMIGA
    IMPORT LONG            pens[16],
                           wordpens[16],
                           longpens[16];
    IMPORT int             divided,
                           fullscreen,
                        // remainder,
                           px, py,
                           size,
                           wide;
    IMPORT UBYTE*          byteptr[MAXBOXHEIGHT];
    IMPORT UWORD*          wordptr[MAXBOXHEIGHT];
    IMPORT ULONG*          longptr[MAXBOXHEIGHT];
    IMPORT struct Catalog* CatalogPtr;
#endif

MODULE int    stopline,
              x,
              y,
              xx,
              yy;
MODULE FLAG   newband,
              skip[FLIPS][4],
              rumble1, rumble2,
              theflag[BOXHEIGHT];  // Flag pin for this line (TRUE or FALSE)
MODULE UBYTE  t,
              bgcolour[BOXHEIGHT], // background colour for this line (taking account of Flag pin) (stored in Arcadia format)
              fgc,                 // stored in Arcadia format
              fgc1, fgc2,          // stored in Arcadia format
              bgc = GREY,          // stored in Arcadia format
              bgc1, bgc2,          // stored in Arcadia format
              hires, // 0 for low resolution, 1 for high resolution
              spritefgc[FLIPS][4][16],
              image[FLIPS][4][8],
              sprcolour,
              sprcolours[BOXWIDTH][BOXHEIGHT],
              boardmode,
              OutputBuffer[7],
              rowbuf[16];
MODULE SLONG  hoffset, // = XMARGIN + 0..7
              spritex[FLIPS][4],
              spritey[FLIPS][4],
              rowc, // goes up only 1 per 2 rasts in low-res mode
                    // 0..7: which y-pixel row in the current "band" (row of chars) we are doing
              row,  // whenever rowc gets to 8, rowc is cleared and row is incremented
              blockmode,
              column,
              thechar,
              tall[FLIPS][4],
              whichsprite,
              imagedata,
              rowlimit,
              flip = 0;

EXPORT       int voffset;
EXPORT const int from_a[8] =
{   0, //  0 ->  0 white
    1, //  1 ->  1 yellow
    4, //  2 ->  4 cyan
    5, //  3 ->  5 green
    2, //  4 ->  2 purple
    3, //  5 ->  3 red
    6, //  6 ->  6 blue
    7, //  7 ->  7 black
};

/* Bottom margins are as follows (for initial screen):
 HOBO:     (very thick) yellow
 THE-END:  thick green, medium yellow, thin cyan
 TURTLES:  (very thick) green
 JUMPBUG?: thin green, thick yellow
These were tested under MESS V2 and WinArcadia 1.5.
*/

// MODULE FUNCTIONS

MODULE void a_playerinput(int source, int dest);
MODULE void drawsprites(SLONG flipper);
MODULE __inline void a_emuinput(void);
MODULE int sprite_collision(int first, int second);
MODULE void drawallsprites(void);
MODULE __inline void collidesprites(void);
MODULE __inline void breakdma(void);
MODULE __inline void breakrastline(void);
MODULE __inline void a_drawpixel(int x, int y, int colour);
MODULE __inline void a_drawline(int y, int colour);

/* start of code

Test version of uvi() which measures graphics rendering speed.
EXPORT __inline void uvi(void)
{   for (bgc = 0; bgc <= 7; bgc++)
    {   for (y = 0; y < BOXHEIGHT; y++)
        {   for (x = 0; x < BOXWIDTH; x++)
            {   a_drawpixel(x, y, bgc);
        }   }
        updatescreen();
        frames++;
}   } */

EXPORT __inline void uvi(void)
{   // assert(machine == ARCADIA);

    inframe = TRUE;
    rumble1 = rumble2 = FALSE;

    for (x = 0; x < BOXWIDTH; x++)
    {   for (y = 0; y < BOXHEIGHT; y++)
        {   sprcolours[x][y] = 0;
    }   }

    voffset = (SWORD) 255 - memory[A_VSCROLL];

    if (memory[A_RESOLUTION] & 0x40) // lower screen on/off flag
    {   rowlimit = 26;
    } else
    {   rowlimit = 13;
    }
    hires = (memory[A_BGCOLOUR] & 0x80) >> 7; // %10000000 -> %00000001 (doublescan flag)

    psu &= ~(PSU_S);
    uviwrite(A_CHARLINE, 0xFF);
    breakdma();

    // top margin---------------------------------------------------------

    if (voffset)
    {   if (voffset < BOXHEIGHT)
        {   stopline = voffset;
        } else
        {   stopline = BOXHEIGHT;
        }
        for (y = 0; y < stopline; y++)
        {   if (flagline && (psu & PSU_F))
            {   theflag[y] = PSU_F;
                bgcolour[y] = bgc = 7 - (memory[A_BGCOLOUR] & 0x07);
            } else
            {   theflag[y] = 0;
                bgcolour[y] = bgc = memory[A_BGCOLOUR] & 0x07;
            }

            slice = CPL - overcalc;
            cpu();
            CHECKINPUT;

            a_drawline(y, bgc); // draw entire (blank) line
    }   }

    uviwrite(A_P1PADDLE,      0xFF); // "The potentiometer is set to $FF
    uviwrite(A_P2PADDLE,      0xFF); // on the trailing edge of VRST."

    // These have to be commented out for DORAEMON collision detection to work reliably
    // uviwrite(A_SPRITECOLLIDE, 0xFF); // "Each [bit] is reset to 1...at the end of the vertical blanking period."
    // uviwrite(A_BGCOLLIDE,     0xFF); // "Each [bit] is reset [to 1]...at the end of the vertical blanking period."

    // main (character) area----------------------------------------------

    rowc     =
    row      = 0;
    newband  = TRUE;
    y        = voffset;

    for (rast = 0; rast < 262; rast++) // NTSC has 262 rastlines, value for true PAL would be 312?
    {   if (newband)
        {   // start of a new "band" (row) of characters (128*8 pixels in size)

            // BOWLING relies on boardmode being refreshed for each new "band"
            boardmode = memory[A_PITCH] & 0x80;
            if (boardmode)
            {   fgc1 = (memory[A_RESOLUTION] & 0x38) >> 3; // %00111000 -> %00000111
                fgc2 = (memory[A_BGCOLOUR]   & 0x38) >> 3; // %00111000 -> %00000111 colours of tile set 0
                bgc1 =  memory[A_RESOLUTION] & 0x07;       // %00000111
                bgc2 =  memory[A_BGCOLOUR]   & 0x07;       // %00000111              background colour
            }

            // REDCLASH relies on the row of characters being cached
            if (row <= 12)
            {   for (x = 0; x < 16; x++)
                {   rowbuf[x] = memory[A_OFFSET_UPPERSCREEN + (row << 4) + x];
            }   }
            elif (row <= 25)
            {   // row can be 26 or more, but rowbuf[] is not used in that case
                for (x = 0; x < 16; x++)
                {   rowbuf[x] = memory[0x1930               + (row << 4) + x];
            }   }

            hoffset = A_XMARGIN + ((memory[A_VOLUME] & 0xE0) >> 5); // 3 bits wide
            row++;

            newband = FALSE;
            if (memory[A_CHARLINE] != dma[(memory[A_BGCOLOUR] & 0x80) >> 7][rast >> 3])
            {   uviwrite(A_CHARLINE, dma[(memory[A_BGCOLOUR] & 0x80) >> 7][rast >> 3]);
                breakdma();
        }   }
        breakrastline();

        if (y < BOXHEIGHT)
        {   if (flagline && (psu & PSU_F))
            {   theflag[y] = PSU_F;
                bgc = bgcolour[y] = 7 - (memory[A_BGCOLOUR] & 0x07);
            } else
            {   theflag[y] = 0;
                bgc = bgcolour[y] =      memory[A_BGCOLOUR] & 0x07 ;
            }

            slice = CPL - overcalc;
            cpu();
            CHECKINPUT;

            // ** draw one Y-line of the display
            if (row <= rowlimit)
            {   // left margin
                for (x = 0; x < hoffset; x++)
                {   a_drawpixel(x, y, bgc);
                }

                // centre area
                // assert(x == hoffset);
                // x = hoffset;
                blockmode = (memory[A_RESOLUTION] & 0x80) >> 7;
                for (column = 0; column < 16; column++)
                {   // ** draw one character (or block graphic) **

                    thechar = rowbuf[column];
                    if (thechar == 0xC0)
                    {   blockmode = 1;
                    } elif (thechar == 0x40)
                    {   blockmode = 0;
                    }

                    // get colours for this tile
                    if (boardmode)
                    {   if (theflag[y]) // or: if (flagline && (psu & PSU_F))
                        {   if (thechar & 0x40)
                            {   fgc = (UBYTE) (7 - fgc2);
                            } else
                            {   fgc = (UBYTE) (7 - fgc1);
                            }
                            if (thechar & 0x80)
                            {   bgc = (UBYTE) (7 - bgc2);
                            } else
                            {   bgc = (UBYTE) (7 - bgc1);
                        }   }
                        else
                        {   if (thechar & 0x40)
                            {   fgc =     fgc2;
                            } else
                            {   fgc =     fgc1;
                            }
                            if (thechar & 0x80)
                            {   bgc =     bgc2;
                            } else
                            {   bgc =     bgc1;
                    }   }   }
                    else
                    {   if (theflag[y]) // or: if (flagline && (psu & PSU_F))
                        {   fgc = 7
                                - ((thechar              &  0xC0) >> 5)  // %11000000 -> %00000110 (0, 2, 4 or 6)
                                - ((memory[A_BGCOLOUR] >>    3) &  1); // 0 or 1
                            bgc = 7 - (memory[A_BGCOLOUR] & 0x07);
                        } else
                        {   fgc = ((thechar              &  0xC0) >> 5)  // %11000000 -> %00000110 (0, 2, 4 or 6)
                                + ((memory[A_BGCOLOUR] >>    3) &  1); // 0 or 1
                            bgc =      memory[A_BGCOLOUR] & 0x07 ;
                    }   }

                    // get image data for this tile
                    thechar &= 0x3F;
                    if (!blockmode && thechar >= 56) // this is a UDC
                    {   imagedata = memory[A_OFFSET_SPRITES + (8 * (thechar - 56)) + rowc];
                    } else // built-in graphics
                    {   imagedata = pix[blockmode][thechar][rowc];
                    }

                    // draw one row of this tile (ie. 8*1)
                    for (xx = 0; xx < 8; xx++)
                    {   if (imagedata & (0x80 >> xx))
                        {   a_drawpixel(x, y, fgc);
                        } else
                        {   a_drawpixel(x, y, bgc);
                        }
                        x++;
                }   }

                // right margin
                // in case we are in board mode
                // assert(x == hoffset + 128);
                do
                {   a_drawpixel(x++, y, bgcolour[y]);
                } while (x < BOXWIDTH);
            } else
            {   // bottom margin

                a_drawline(y, bgc); // draw entire (blank) line
        }   }
        else // this raster is not shown
        {   slice = CPL - overcalc;
            cpu();
            CHECKINPUT;
        }

        if (rast == INPUTLINE) // 207
        {   a_emuinput();
        } elif (rast == SENSELINE) // 240
        {   psu |= PSU_S;
        }

        if (hires || (rast & 1))
        {   if (rowc == 7)
            {   rowc = 0;
                newband = TRUE;
            } else rowc++;
        }

        y++;
    }

    // rastline is 0 (262), Sense bit is set, CHARLINE is $FD
    drawallsprites(); // must not be done at SENSELINE, eg. REDCLASH
    collidesprites(); // must not be done at SENSELINE, eg. ROBOTKIL, ESCAPE

    psu &= ~(PSU_S); // clear Sense bit. Necessary, eg. CIRCUS
    uviwrite(A_CHARLINE, 0xFF); // possibly unnecessary or wrong?
    breakdma();

    // actual diagonal (bottom-right to top-left) beam retrace is here
    slice = (CPL * 3) - overcalc;
    cpu();
    CHECKINPUT;

/*  if (rumble)
    {   if (rumble1)
        {   for (y = 0; y < BOXHEIGHT; y++)
            {   a_drawpixel(0, y, 7 - screen[0][y]);
        }   }
        if (rumble2)
        {   for (y = 0; y < BOXHEIGHT; y++)
            {   a_drawpixel(BOXWIDTH - 1, y, 7 - screen[BOXWIDTH - 1][y]);
    }   }   } */

    // Currently, frame skipping only affects warp mode when "limit
    // refresh rate" is off.

    if
    (   (!warp || !limitrefresh)
     && (frames % frameskip == 0)
    )
    {   updatescreen();
    }

    flip ^= 1; // means: if (flip) flip = 0; else flip = 1;
    frames++;

    inframe = FALSE;
    if (crippled)
    {   uncripple();
    }

    if (recmode == RECMODE_PLAY && offset >= recsize)
    {   if (loop)
        {   macro_restartplayback();
        } else
        {   macro_stop();
    }   }

    if (runtoframe)
    {   runtoframe = FALSE;
        T1;
        printf
        (   GetCatalogStr(CatalogPtr,
                MSG_XVI_REACHEDFRAME,
                "Reached next frame (%ld).\n\n"
            ),
            (long int) frames
        );
        T2;
        pause();
}   }

MODULE void drawallsprites(void)
{   // cache sprite variables
       tall[flip][0] = 2 - ((memory[A_SPRITES01CTRL] & 0x80) >> 7); // %10000000 -> %00000001
       tall[flip][1] = 2 - ((memory[A_SPRITES01CTRL] & 0x40) >> 6); // %01000000 -> %00000001
       tall[flip][2] = 2 - ((memory[A_SPRITES23CTRL] & 0x80) >> 7); // %10000000 -> %00000001
       tall[flip][3] = 2 - ((memory[A_SPRITES23CTRL] & 0x40) >> 6); // %01000000 -> %00000001
    spritex[flip][0] =               memory[A_SPRITE0X] - 43 + A_XMARGIN;
    spritey[flip][0] = (SWORD) 256 - memory[A_SPRITE0Y];
    spritex[flip][1] =               memory[A_SPRITE1X] - 43 + A_XMARGIN;
    spritey[flip][1] = (SWORD) 256 - memory[A_SPRITE1Y];
    spritex[flip][2] =               memory[A_SPRITE2X] - 43 + A_XMARGIN;
    spritey[flip][2] = (SWORD) 256 - memory[A_SPRITE2Y];
    spritex[flip][3] =               memory[A_SPRITE3X] - 43 + A_XMARGIN;
    spritey[flip][3] = (SWORD) 256 - memory[A_SPRITE3Y];

    t = 0xFF;
    for (whichsprite = 0; whichsprite < 4; whichsprite++)
    {   // cache sprite background flag pin state
        for (y = 0; y < tall[flip][whichsprite] * 8; y++)
        {   yy = spritey[flip][whichsprite] + y;
            if (yy >= BOXHEIGHT)
            {   break;
            }
            switch(whichsprite)
            {
            case 0:
                if (theflag[yy])
                {   spritefgc[flip][whichsprite][y] = (UBYTE) 7 - ((memory[A_SPRITES01CTRL] & 0x38) >> 3); // %00111000 -> %00000111
                } else
                {   spritefgc[flip][whichsprite][y] = (UBYTE)     ((memory[A_SPRITES01CTRL] & 0x38) >> 3); // %00111000 -> %00000111
                }
            acase 1:
                if (theflag[yy])
                {   spritefgc[flip][whichsprite][y] = (UBYTE) 7 - ( memory[A_SPRITES01CTRL] & 0x07      ); // %00000111 -> %00000111
                } else
                {   spritefgc[flip][whichsprite][y] = (UBYTE)     ( memory[A_SPRITES01CTRL] & 0x07      ); // %00000111 -> %00000111
                }
            acase 2:
                if (theflag[yy])
                {   spritefgc[flip][whichsprite][y] = (UBYTE) 7 - ((memory[A_SPRITES23CTRL] & 0x38) >> 3); // %00111000 -> %00000111
                } else
                {   spritefgc[flip][whichsprite][y] = (UBYTE)     ((memory[A_SPRITES23CTRL] & 0x38) >> 3); // %00111000 -> %00000111
                }
            acase 3:
                if (theflag[yy])
                {   spritefgc[flip][whichsprite][y] = (UBYTE) 7 - ( memory[A_SPRITES23CTRL] & 0x07      ); // %00000111 -> %00000111
                } else
                {   spritefgc[flip][whichsprite][y] = (UBYTE)     ( memory[A_SPRITES23CTRL] & 0x07      ); // %00000111 -> %00000111
                }
            adefault:
                // assert(0);
            break;
        }   }

        // cache sprite imagery
        for (y = 0; y < 8; y++)
        {   image[flip][whichsprite][y] = memory[A_OFFSET_SPRITES + (whichsprite * 8) + y];
        }

/* bgcolour[y] is sampled (written) at the start of the right margin...
   but really we could have several bgcolours for the one rastline, so the below reliance on
   bgcolour[y] is somewhat dodgy... */

        // check sprite position
        if (memory[A_SPRITE0X + (2 * whichsprite)] > 227)
        {   // "If the HC is set to >227, the object is effectively
            // removed from the video field."
            skip[flip][whichsprite] = TRUE;
        } else
        {   skip[flip][whichsprite] = FALSE;

            // detect sprite-background collisions
            // it is important that this is done BEFORE drawing any sprites
            if (collisions)
            {   x = spritex[flip][whichsprite];
                y = spritey[flip][whichsprite];
                if (tall[flip][whichsprite] == 1)
                {   for (xx = 0; xx < 8; xx++)
                    {   if (x >= 0 && x < BOXWIDTH)
                        {   for (yy = 0; yy < 8; yy++)
                            {   if (y >= 0 && y < BOXHEIGHT)
                                {   imagedata = image[flip][whichsprite][yy];

                                    if
                                    (   (imagedata & (0x80 >> xx)) // if this sprite-pixel is set
                                     && screen[x][y] != from_a[bgcolour[y]]
                                    )
                                    {   t &= ~(1 << whichsprite);
                                        if (p1sprite == whichsprite) rumble1 = TRUE;
                                        if (p2sprite == whichsprite) rumble2 = TRUE;
                                        break;
                                    }
                                }
                                y++;
                            }
                            y -= 8;
                        }
                        x++;
                }   }
                else
                {   // assert(tall[flip][whichsprite] == 2);
                    for (xx = 0; xx < 8; xx++)
                    {   if (x >= 0 && x < BOXWIDTH)
                        {   for (yy = 0; yy < 8; yy++)
                            {   if (y >= 0 && y < BOXHEIGHT)
                                {   imagedata = image[flip][whichsprite][yy];

                                    if (imagedata & (0x80 >> xx)) // if this sprite-pixel is set
                                    {   if (screen[x][y] != from_a[bgcolour[y]])
                                        {   t &= ~(1 << whichsprite);
                                            if (p1sprite == whichsprite) rumble1 = TRUE;
                                            if (p2sprite == whichsprite) rumble2 = TRUE;
                                            break;
                                        }
                                        if (y < BOXHEIGHT - 1)
                                        {   if (screen[x][y + 1] != from_a[bgcolour[y + 1]])
                                            {   t &= ~(1 << whichsprite);
                                                if (p1sprite == whichsprite) rumble1 = TRUE;
                                                if (p2sprite == whichsprite) rumble2 = TRUE;
                                                break;
                                }   }   }   }
                                y += 2;
                            }
                            y -= 16;
                        }
                        x++;
    }   }   }   }   }
    uviwrite(A_BGCOLLIDE, t);

    if (frames && demultiplex) // if first frame, we won't have valid data! :-(
    {   if (flip)
        {   drawsprites(0);
        } else
        {   drawsprites(1);
    }   }
    drawsprites(flip);
}

MODULE __inline void collidesprites(void)
{   /* This doesn't use screen[], thus it is safe at any point in the frame. */

    t = 0xFF;
    if (collisions)
    {   if (sprite_collision(0, 1))
        {   t &=  ~1; // AND with %11111110 (clear bit 0)
            if (p1sprite == 0 || p1sprite == 1) rumble1 = TRUE;
            if (p2sprite == 0 || p2sprite == 1) rumble2 = TRUE;
        }
        if (sprite_collision(0, 2))
        {   t &=  ~2; // AND with %11111101 (clear bit 1)
            if (p1sprite == 0 || p1sprite == 2) rumble1 = TRUE;
            if (p2sprite == 0 || p2sprite == 2) rumble2 = TRUE;
        }
        if (sprite_collision(0, 3))
        {   t &=  ~4; // AND with %11111011 (clear bit 2)
            if (p1sprite == 0 || p1sprite == 3) rumble1 = TRUE;
            if (p2sprite == 0 || p2sprite == 3) rumble2 = TRUE;
        }
        if (sprite_collision(1, 2))
        {   t &=  ~8; // AND with %11110111 (clear bit 3)
            if (p1sprite == 1 || p1sprite == 2) rumble1 = TRUE;
            if (p2sprite == 1 || p2sprite == 2) rumble2 = TRUE;
        }
        if (sprite_collision(1, 3))
        {   t &= ~16; // AND with %11101111 (clear bit 4)
            if (p1sprite == 1 || p1sprite == 3) rumble1 = TRUE;
            if (p2sprite == 1 || p2sprite == 3) rumble2 = TRUE;
        }
        if (sprite_collision(2, 3))
        {   t &= ~32; // AND with %11011111 (clear bit 5)
            if (p1sprite == 2 || p1sprite == 3) rumble1 = TRUE;
            if (p2sprite == 2 || p2sprite == 3) rumble2 = TRUE;
    }   }
    uviwrite(A_SPRITECOLLIDE, t);
}

MODULE void drawsprites(SLONG flipper)
{   for (whichsprite = 0; whichsprite < 4; whichsprite++)
    {   if (!skip[flipper][whichsprite])
        {   x = spritex[flipper][whichsprite];
            y = spritey[flipper][whichsprite];

            if (demultiplex && flipper != flip)
            {   // we are doing an "old" sprite (ie. one of the 4 from last frame)
                // check its position against that of its "new" sprite equivalent
                // if the rectangles intersect, suppress display of the "old" sprite
                // this is to avoid "smearing" when demultiplexing

                // check whether they are too close horizontally
                if
                (   x > spritex[flip][whichsprite] - 8
                 && x < spritex[flip][whichsprite] + 8
                )
                {   // check whether they are too close vertically
                    if
                    (   y + (8 * tall[flipper][whichsprite]) > spritey[flip][whichsprite]
                        // if the bottommost pixel of the "old" sprite is above the topmost pixel of the "new" sprite
                     && y < spritey[flip][whichsprite] + (8 * tall[flip][whichsprite])
                        // or the topmost pixel of the "old" sprite is below the bottommost pixel of the "new" sprite
                    )
                    // could do skip[flipper][whichsprite] = TRUE; but unnecessary
                    continue;
            }   }

            if (tall[flipper][whichsprite] == 1)
            {   for (yy = 0; yy < 8; yy++)
                {   if (y >= 0 && y < BOXHEIGHT)
                    {   imagedata =     image[flipper][whichsprite][yy];
                        sprcolour = spritefgc[flipper][whichsprite][yy];

                        for (xx = 0; xx < 8; xx++)
                        {   if (imagedata & (0x80 >> xx)) // if this x-bit is set
                            {   if (x >= 0 && x < BOXWIDTH)
                                {   sprcolours[x][y] |= sprcolour;
                                    a_drawpixel(x, y, sprcolours[x][y]);
                            }   }
                            x++;
                        }
                        x -= 8;
                    }
                    y++;
            }   }
            else
            {   for (yy = 0; yy < 8; yy++) // for each y-line of the sprite
                {   if (y >= 0 && y < BOXHEIGHT)
                    {   imagedata =     image[flipper][whichsprite][yy];
                        sprcolour = spritefgc[flipper][whichsprite][yy * 2];

                        for (xx = 0; xx < 8; xx++)
                        {   if (imagedata & (0x80 >> xx)) // if this x-bit is set
                            {   if (x >= 0 && x < BOXWIDTH)
                                {   sprcolours[x][y] |= sprcolour;
                                    a_drawpixel(x, y, sprcolours[x][y]);

                                    if (y < BOXHEIGHT - 1)
                                    {   sprcolours[x][y + 1] |= sprcolour;
                                        a_drawpixel(x, y + 1, sprcolours[x][y + 1]);
                            }   }   }
                            x++;
                        } // rof (end for)
                        x -= 8;
                    }
                    y += 2;
}   }   }   }   }

/* Each raster is about  12x 16  =  192 clock periods.
   Each frame  is about 192x262/2       clock periods.
*/

MODULE void a_playerinput(int source, int dest)
{   // dest is which side   (0 or 1) you want to set the registers of.
    // source is which side (0 or 1) you want to use to do it.

    if (recmode == RECMODE_PLAY)
    {   t = IOBuffer[offset++];
        uviwrite((UWORD) (A_P1PADDLE     -  dest     ),           t              );
        t = IOBuffer[offset++];
        uviwrite((UWORD) (A_P1LEFTKEYS   + (dest * 4)), (UBYTE) ((t & 0xF0) >> 4)); // bits 7..4 -> bits  3..0
        uviwrite((UWORD) (A_P1MIDDLEKEYS + (dest * 4)), (UBYTE)  (t & 0x0F)      ); // bits 3..0 -> bits  3..0 (masking is indeed necessary)
        t = IOBuffer[offset++];
        uviwrite((UWORD) (A_P1RIGHTKEYS  + (dest * 4)), (UBYTE) ((t & 0xF0) >> 4)); // bits 7..4 -> bits  3..0
        uviwrite((UWORD) (A_P1PALLADIUM  + (dest * 4)), (UBYTE)  (t & 0x0F)      ); // bits 3..0 -> bits  3..0 (masking is indeed necessary)
    } else
    {   // assert(recmode == RECMODE_NORMAL || recmode == RECMODE_RECORD);

        if
        (   (dest == 0 && connected == NET_CLIENT)
         || (dest == 1 && connected == NET_SERVER)
        )
        {   return;
        }

        joystate = ReadJoystick((UWORD) (1 - source));

        memory[A_P1LEFTKEYS   + (dest * 4)] =
        memory[A_P1MIDDLEKEYS + (dest * 4)] =
        memory[A_P1RIGHTKEYS  + (dest * 4)] =
        memory[A_P1PALLADIUM  + (dest * 4)] = 0;
        if (autofire)
        {   if
            (   (!requirebutton || (joystate & JOYFIRE1) || KeyDown(keypads[source][0]))
             && ((frames % totalframes) < downframes)
            )
            {   // 1st firebutton
                memory[keytable[key1].a_address + (dest * 4)] |= keytable[key1].a_mask;
        }   }
        else
        {   if ((joystate & JOYFIRE1) || KeyDown(keypads[source][0]))
            {   // 1st firebutton
                memory[keytable[key1].a_address + (dest * 4)] |= keytable[key1].a_mask;
        }   }

        if
        (   (joystate & JOYFIRE2)
         || (KeyDown(keypads[source][21]))
        )
        {   // 2nd firebutton
            memory[keytable[key2].a_address + (dest * 4)] |= keytable[key2].a_mask;
        }
        if
        (   (joystate & JOYFIRE3)
         || (KeyDown(keypads[source][22]))
        )
        {   // 3rd firebutton
            memory[keytable[key3].a_address + (dest * 4)] |= keytable[key3].a_mask;
        }
        if
        (   (joystate & JOYFIRE4)
         || (KeyDown(keypads[source][23]))
        )
        {   // 4th firebutton
            memory[keytable[key4].a_address + (dest * 4)] |= keytable[key4].a_mask;
        }

        // left column
        t = memory[A_P1LEFTKEYS + (dest * 4)];
        t |= (0x08 * KeyDown(keypads[source][ 1])); // "1"  key (Arcadia)
        t |= (0x04 * KeyDown(keypads[source][ 4])); // "4"  key (Arcadia)
        t |= (0x02 * KeyDown(keypads[source][ 7])); // "7"  key (Arcadia)
        t |= (0x01 * KeyDown(keypads[source][10])); // "Cl" key (Arcadia)
        uviwrite((UWORD) (A_P1LEFTKEYS   + (dest * 4)), t);

        // middle column
        t = memory[A_P1MIDDLEKEYS + (dest * 4)];
        t |= (0x08 * KeyDown(keypads[source][ 2])); // "2"  key (Arcadia)
        t |= (0x04 * KeyDown(keypads[source][ 5])); // "5"  key (Arcadia)
        t |= (0x02 * KeyDown(keypads[source][ 8])); // "8"  key (Arcadia)
        t |= (0x01 * KeyDown(keypads[source][11])); // "0"  key (Arcadia)
        uviwrite((UWORD) (A_P1MIDDLEKEYS + (dest * 4)), t);

        // right column
        t = memory[A_P1RIGHTKEYS + (dest * 4)];
        t |= (0x08 * KeyDown(keypads[source][ 3])); // "3"  key (Arcadia)
        t |= (0x04 * KeyDown(keypads[source][ 6])); // "6"  key (Arcadia)
        t |= (0x02 * KeyDown(keypads[source][ 9])); // "9"  key (Arcadia)
        t |= (0x01 * KeyDown(keypads[source][12])); // "En" key (Arcadia)
        uviwrite((UWORD) (A_P1RIGHTKEYS  + (dest * 4)), t);

        // Palladium "column"
        t = memory[A_P1PALLADIUM + (dest * 4)];
        t |= (0x08 * KeyDown(keypads[source][16])); // "x4" key (Arcadia)
        t |= (0x04 * KeyDown(keypads[source][15])); // "x3" key (Arcadia)
        t |= (0x02 * KeyDown(keypads[source][14])); // "x2" key (Arcadia)
        t |= (0x01 * KeyDown(keypads[source][13])); // "x1" key (Arcadia)
        uviwrite((UWORD) (A_P1PALLADIUM + (dest * 4)), t);

        dopaddle(source, dest);

        if (memory[A_BGCOLOUR] & 0x40) // paddle interpolation bit
        {   uviwrite((UWORD) (A_P1PADDLE - dest), (UBYTE) ax[dest]);
        } else
        {   uviwrite((UWORD) (A_P1PADDLE - dest), (UBYTE) ay[dest]);
    }   }

    if (memory[A_BGCOLOUR] & 0x40) // paddle interpolation bit
    {   sx[dest] = (UBYTE) ax[dest];
    } else
    {   sy[dest] = (UBYTE) ay[dest];
}   }

MODULE __inline void a_emuinput(void)
{   UBYTE therecv[4],
          thesend[4];

    readkybd();

    // must always do a_playerinput(foo, 0) then a_playerinput(foo, 1).

    if (swapped)
    {   a_playerinput(1, 0);
        a_playerinput(0, 1);
    } else
    {   a_playerinput(0, 0);
        a_playerinput(1, 1);
    }

    if (recmode != RECMODE_PLAY)
    {   // assert(recmode == RECMODE_NORMAL || recmode == RECMODE_RECORD);

        joystate = ReadJoystick(1) | ReadJoystick(0);

        if (KeyDown(console[0]) || (joystate & JOYSTART))
        {   t = 1;
        } else
        {   t = 0;
        }
        if (KeyDown(console[1]) || (joystate & JOYA    )) { t |= 2; }
        if (KeyDown(console[2]) || (joystate & JOYB    )) { t |= 4; }

        uviwrite(A_CONSOLE, t);

        domouse();
    }

    /* 1st argument is pointer to receive buffer.
       2nd argument is pointer to send buffer.
       3rd argument is size in bytes of each buffer

    Server is left player, client is right player.
    4 bytes per frame (27 significant bits) are exchanged for Arcadia.
    3 bytes per frame (24 significant bits) are exchanged for Interton/Elektor. */

    if (connected == NET_SERVER)
    {   // we are left player

        thesend[0] = ((memory[A_P1LEFTKEYS  ] & 0x0F) <<  4) // bits 3..0 -> bits  7..4 (masking not really needed)
                   |  (memory[A_P1MIDDLEKEYS] & 0x0F);       // bits 3..0 -> bits  3..0 (masking not really needed)
        thesend[1] = ((memory[A_P1RIGHTKEYS ] & 0x0F) <<  4) // bits 3..0 -> bits  7..4 (masking not really needed)
                   |  (memory[A_P1PALLADIUM ] & 0x0F);       // bits 3..0 -> bits  3..0 (masking not really needed)
        thesend[2] =   memory[A_P1PADDLE];                   // bits 7..0 -> bits  7..0
        thesend[3] =   memory[A_CONSOLE];                    // bits 7..0 -> bits  7..0

        if (NetExchange((char *) &therecv, (const char *) &thesend, 4))
        {   uviwrite(A_P2LEFTKEYS  , (UBYTE) ((therecv[0] & 0xF0) >> 4)); // bits 7..4 -> bits 3..0
            uviwrite(A_P2MIDDLEKEYS, (UBYTE)  (therecv[0] & 0x0F)      ); // bits 3..0 -> bits 3..0
            uviwrite(A_P2RIGHTKEYS , (UBYTE) ((therecv[1] & 0xF0) >> 4)); // bits 7..4 -> bits 3..0
            uviwrite(A_P2PALLADIUM , (UBYTE)  (therecv[1] & 0x0F)      ); // bits 3..0 -> bits 3..0
            uviwrite(A_P2PADDLE    ,           therecv[2]              ); // bits 7..0 -> bits 7..0
            uviwrite(A_CONSOLE     , (UBYTE)  (therecv[3] | memory[A_CONSOLE]));
    }   }
    elif (connected == NET_CLIENT)
    {   // we are right player

        thesend[0] = ((memory[A_P2LEFTKEYS  ] & 0x0F) <<  4) // bits 3..0 -> bits  7..4 (masking not really needed)
                   |  (memory[A_P2MIDDLEKEYS] & 0x0F);       // bits 3..0 -> bits  3..0 (masking not really needed)
        thesend[1] = ((memory[A_P2RIGHTKEYS ] & 0x0F) <<  4) // bits 3..0 -> bits  7..4 (masking not really needed)
                   |  (memory[A_P2PALLADIUM ] & 0x0F);       // bits 3..0 -> bits  3..0 (masking not really needed)
        thesend[2] =   memory[A_P2PADDLE];                   // bits 7..0 -> bits  7..0
        thesend[3] =   memory[A_CONSOLE];                    // bits 7..0 -> bits  7..0

        if (NetExchange((char *) &therecv, (const char *) &thesend, 4))
        {   uviwrite(A_P1LEFTKEYS  , (UBYTE) ((therecv[0] & 0xF0) >> 4)); // bits 7..4 -> bits 3..0
            uviwrite(A_P1MIDDLEKEYS, (UBYTE)  (therecv[0] & 0x0F)      ); // bits 3..0 -> bits 3..0
            uviwrite(A_P1RIGHTKEYS , (UBYTE) ((therecv[1] & 0xF0) >> 4)); // bits 7..4 -> bits 3..0
            uviwrite(A_P1PALLADIUM , (UBYTE)  (therecv[1] & 0x0F)      ); // bits 3..0 -> bits 3..0
            uviwrite(A_P1PADDLE    ,           therecv[2]              ); // bits 7..0 -> bits 7..0
            uviwrite(A_CONSOLE     , (UBYTE)  (therecv[3] | memory[A_CONSOLE]));
    }   }

    if (recmode == RECMODE_PLAY)
    {   uviwrite(A_CONSOLE, IOBuffer[offset++]);
    } elif (recmode == RECMODE_RECORD)
    {   // assert(MacroHandle);

        OutputBuffer[0] =   memory[A_P1PADDLE];                   // bits 7..0 -> bits  7..0
        OutputBuffer[1] = ((memory[A_P1LEFTKEYS  ] & 0x0F) <<  4) // bits 3..0 -> bits  7..4 (masking not really needed)
                        |  (memory[A_P1MIDDLEKEYS] & 0x0F);       // bits 3..0 -> bits  3..0 (masking not really needed)
        OutputBuffer[2] = ((memory[A_P1RIGHTKEYS ] & 0x0F) <<  4) // bits 3..0 -> bits  7..4 (masking not really needed)
                        |  (memory[A_P1PALLADIUM ] & 0x0F);       // bits 3..0 -> bits  3..0 (masking not really needed)
        OutputBuffer[3] =   memory[A_P2PADDLE];                   // bits 7..0 -> bits  7..0
        OutputBuffer[4] = ((memory[A_P2LEFTKEYS  ] & 0x0F) <<  4) // bits 3..0 -> bits  7..4 (masking not really needed)
                        |  (memory[A_P2MIDDLEKEYS] & 0x0F);       // bits 3..0 -> bits  3..0 (masking not really needed)
        OutputBuffer[5] = ((memory[A_P2RIGHTKEYS ] & 0x0F) <<  4) // bits 3..0 -> bits  7..4 (masking not really needed)
                        |  (memory[A_P2PALLADIUM ] & 0x0F);       // bits 3..0 -> bits  3..0 (masking not really needed)
        OutputBuffer[6] =   memory[A_CONSOLE];                    // bits 7..0 -> bits  7..0 (5 bits are wasted)

        DISCARD fwrite(OutputBuffer, 7, 1, MacroHandle);
        // should really check return code of fwrite()
        // 7 bytes per frame * 60 frames per second = 420 bytes/sec.
}   }

MODULE int sprite_collision(int first, int second)
{   int        b1,
               b2,
               y;
    signed int x;

    // check whether they are close enough
    if
    (   spritex[flip][first] <= spritex[flip][second] - 8
     || spritex[flip][first] >= spritex[flip][second] + 8
    )
    {   return FALSE;
    }

    // Undrawn sprites are not collision-checked! This behaviour
    // is needed for eg. DORAEMON.
    if (skip[flip][first] || skip[flip][second])
    {   return FALSE;
    }

    // calculate how much further to the right the first is
    // than the second, eg.
    // x <  0: first  is left, second is right
    // x == 0: both are at same X-coordinate
    // x >  0: second is left, first  is right
    x = (signed int) (spritex[flip][first] - spritex[flip][second]);

    for (y = 0; y < 8; y++)
    {   // check whether we are above the top of the second sprite
        if
        (   spritey[flip][first ] + (y * tall[flip][first ])
         <  spritey[flip][second]
        )
        {   continue;
        }
        // check whether we are below the bottom of the second sprite
        if
        (   spritey[flip][first ] + (y * tall[flip][first ])
         >= spritey[flip][second] + (8 * tall[flip][second])
        )
        {   return FALSE;
        }

/*      Examples:

        195 00000000
        196 11111111
        197 22222222
        198 33333333
        199 44444444
        200 55555555    00000000
        201 66666666    00000000
        202 77777777    11111111
        203             11111111
        204             22222222
        205             22222222 etc.

        yy would be -5.

        195 00000000
        196 00000000
        197 11111111
        198 11111111
        199 22222222
        200 22222222    00000000
        201 33333333    11111111
        202 33333333    22222222
        203 44444444    33333333
        204 44444444    44444444
        205 55555555    55555555 etc.

        yy would be -5.

        195             00000000
        196             11111111
        197             22222222
        198             33333333
        199             44444444
        200 00000000    55555555
        201 00000000    66666666
        202 11111111    77777777
        203 11111111
        204 22222222
        205 22222222 etc.

        yy would be 5.

        195             00000000
        196             00000000
        197             11111111
        198             11111111
        199             22222222
        200 00000000    22222222
        201 00000000    33333333
        202 11111111    33333333
        203 11111111    44444444
        204 22222222    44444444
        205 22222222    55555555 etc.

        yy would be 5. */

        yy = spritey[flip][first] - spritey[flip][second];
        b1 = image[flip][first][y];

        if (tall[flip][first] == tall[flip][second]) // same vs. same
        {   b2 = image[flip][second][yy + y];
        } elif (tall[flip][first] == 1 && tall[flip][second] == 2) // short vs. tall
        {   b2 = image[flip][second][(yy + y) / 2];
        } else
        {   // assert(tall[flip][first] == 2 && tall[flip][second] == 1); // tall vs. short
            b2 = image[flip][second][yy + (y / 2)];
        }

        if (x < 0) b2 >>= -x;
        if (x > 0) b1 >>= x;
        if (b1 & b2)
        {   return TRUE;
    }   }

    return FALSE;
}

MODULE __inline void breakdma(void)
{   if (runtodma)
    {   runtodma = FALSE;
        T1;
        printf
        (   GetCatalogStr(CatalogPtr,
                MSG_UVI_REACHEDDMA,
                "Reached next DMA ($%X, at raster line %ld).\n\n"
            ),
            memory[A_CHARLINE],
            (long int) rast
        );
        T2;
        runto();
}   }

MODULE __inline void breakrastline(void)
{   if (runtorastline)
    {   runtorastline = FALSE;
        T1;
        printf
        (   GetCatalogStr(CatalogPtr,
                MSG_XVI_REACHEDRASTLINE,
                "Reached next raster line (%ld).\n\n"
            ),
            (long int) rast
        );
        T2;
        runto();
    }
    if (rast == rastn)
    {   rastn = -1;
        T1;
        printf
        (   GetCatalogStr(CatalogPtr,
                MSG_XVI_REACHEDRASTLINEN,
                "Reached raster line %ld.\n\n"
            ),
            (long int) rast
        );
        T2;
        runto();
}   }

EXPORT void uviwrite(UWORD address, UBYTE data)
{   // assert(address < 32768);

    /* Writes to memory from the UVI always go through this
    routine (so that watchpoints work properly).
      No mirroring is done here, the UVI module is expected to use the
    resolved address.
      If you have a watchpoint on a mirrored address, eg. $3000,
    then ideally an access of $1000 by the UVI should trigger the.
    watchpoint. The easiest way to deal with this would be to resolve
    the address early (during the "set watchpoint" requester. The same
    applies, of course, for the PVI. */

    if (address == wp && data != memory[wp])
    {   getfriendly(wp);
        T1;
        printf
        (   GetCatalogStr(CatalogPtr,
                MSG_UVI_HITWP,
                "UVI is writing $%X to %s at raster %ld!\n\n"
            ),
                       data,
            (char *)   friendly,
            (long int) rast
        );
        T2;
        runto();
    }

    memory[address] = data;
}

MODULE __inline void a_drawpixel(int x, int y, int colour)
{   // assert(x >= 0);
    // assert(x < BOXWIDTH);
    // assert(y >= 0);
    // assert(y < BOXHEIGHT);
    // assert(colour >= 0);
    // assert(colour <= 7);

    colour = from_a[colour];
    if (screen[x][y] == colour)
    {   return;
    }
    screen[x][y] = colour;

#ifdef WIN32
    display[x + (y * BOXWIDTH)] = pens[colour];
#endif
#ifdef AMIGA
    if (fullscreen) // 2*1
    {   *(wordptr[y] + x) = wordpens[colour];
    } else
    {   switch(size)
        {
        case 1:
            if (wide == 1) // 1*1
            {   *(byteptr[y] + x) = pens[colour]; // 1,1
            } else
            {   // assert(wide == 2); // 2*1

                *(wordptr[y] + x) = wordpens[colour];
            }
        acase 2:
            py = y << 1; // py = y * 2;

            if (wide == 1) // 2*2
            {   *(wordptr[py    ] + x) =
                *(wordptr[py + 1] + x) = wordpens[colour];
            } else
            {   // assert(wide == 2); // 4*2

                *(longptr[py    ] + x) =
                *(longptr[py + 1] + x) = longpens[colour];
            }
        adefault:
            // assert(size == 3);

            px = x * 3;
            py = y * 3;

            if (wide == 1) // 3*3
            {   *(byteptr[py    ] + px    ) =
                *(byteptr[py + 1] + px    ) =
                *(byteptr[py + 2] + px    ) =
                *(byteptr[py    ] + px + 1) =
                *(byteptr[py + 1] + px + 1) =
                *(byteptr[py + 2] + px + 1) =
                *(byteptr[py    ] + px + 2) =
                *(byteptr[py + 1] + px + 2) =
                *(byteptr[py + 2] + px + 2) = pens[colour];
            } else
            {   // assert(wide == 2); // 6*3

                *(wordptr[py    ] + px    ) =
                *(wordptr[py + 1] + px    ) =
                *(wordptr[py + 2] + px    ) =
                *(wordptr[py    ] + px + 1) =
                *(wordptr[py + 1] + px + 1) =
                *(wordptr[py + 2] + px + 1) =
                *(wordptr[py    ] + px + 2) =
                *(wordptr[py + 1] + px + 2) =
                *(wordptr[py + 2] + px + 2) = wordpens[colour];
            }
        break;
    }   }

#endif

}

/* Note that ALIENINV uses sprite-background collision detection,
not sprite-sprite collision detection, for the UFO (aka mothership).
The only sprite-sprite collisions are for players colliding with the
enemy bombs.

CIRCUS waits until CHARLINE == $FD, then immediately checks
SPRITECOLLIDE.

PLEIADES doesn't seem to use the SPRITECOLLIDE or BGCOLLIDE registers
at all. Thus, turning off collision detection will have no effect on
this game.

Games with twinkling starfields:
    SATTACK?
    SVULTURE
    THE-END
and probably others. */

MODULE __inline void a_drawline(int y, int colour)
{
#ifdef WIN32
    PERSIST int end; // PERSISTent for speed
#endif

    // assert(y >= 0);
    // assert(y < BOXHEIGHT);
    // assert(colour >= 0);
    // assert(colour <= 7);

    colour = from_a[colour];
    for (x = 0; x < BOXWIDTH; x++)
    {   screen[x][y] = colour;
    }

#ifdef WIN32
    end = (y + 1) * BOXWIDTH;
    for (x = y * BOXWIDTH; x < end; x++)
    {   display[x] = pens[colour];
    }
#endif
#ifdef AMIGA
    if (!fullscreen)
    {   y *= size;
    }

    for (x = 0; x < divided; x++)
    {   *(longptr[y] + x) = longpens[colour];
    }
    if (!fullscreen && size >= 2)
    {   y++;
        for (x = 0; x < divided; x++)
        {   *(longptr[y] + x) = longpens[colour];
        }
        if (size >= 3)
        {   y++;
            for (x = 0; x < divided; x++)
            {   *(longptr[y] + x) = longpens[colour];
    }   }   }

    /* assert(!remainder);
    if (remainder)
    {   for (x = 0; x < remainder; x++)
        {   *(byteptr[y] + (divided << 2) + x) = pens[colour];
    }   } */
#endif
}

